/* This file is concerned with allocating and freeing pages in the physical
 * memory on behalf of the FORK and EXEC system calls.  It also handles
 * allocating and freeing blocks of data on the swap space.  Third, it
 * does the main work concerning swapping pages in and out. The size
 * of a block on the swap device equals to the size of one physical page.
 * The key data structure used is an array of bits each indicating whether
 * the page is free or not. There exists a second array of the same structure
 * for the disk blocks of the swap device.
 * Furthermore there exists three page tables for each process. One for
 * the text-, data- and bss segment. Each page table contains an array
 * of words, one for each page in the virtual address space. Each word
 * contains information about that particular page: whether it's swapped
 * in or out, locked, modified (some attributes may not be supported yet).
 * If the page is in the memory the word contains the physical pagenumber,
 * too.  If the page is swapped out, the word contains the blocknumber on
 * the disk (swap device).
 *
 * The entry points into this file are:
 *   mem_init:	initialize the tables when MM start up
 *   max_hole:	returns the amount of total available memory space
 *   mem_left:	returns the amount of free core memory
 *   sys_mem :  send a message to sys_task
 */

#include "mm.h"
#include <stdarg.h>
#include <minix/com.h>

static mem_info mem;


/*===========================================================================*
 *				mem_init				     *
 *===========================================================================*/
PUBLIC void mem_init()
{
  /* nothing is to be done here since all is done in page_task */
}


/*===========================================================================*
 *				max_hole				     *
 *===========================================================================*/
PUBLIC vir_bytes max_hole()
{
  /* get total free memory including swap device */
  sys_mem(PG_MEM_INFO, &mem);
  return(((vir_bytes)mem.free_swappg + (vir_bytes)mem.free_mempg) * pagesize);
}


/*===========================================================================*
 *				mem_left				     *
 *===========================================================================*/
PUBLIC vir_bytes mem_left()
{
  /* get free memory in core */
  sys_mem(PG_MEM_INFO, &mem);
  return(mem.free_mempg * mem.pagesize);
}


/*===========================================================================*
 *				sys_mem					     *
 *===========================================================================*/
PUBLIC int sys_mem(what, ...)
int what;
{
  message m;
  va_list va;

  va_start(va, what);
  m.m_type = what;
  m.PAR1 = va_arg(va, int);
  m.PAR2 = va_arg(va, int);
  m.PAR3 = va_arg(va, int);
  if (sendrec(PAGER, &m) != OK)
	panic("Kernel didn't respond to sys_mem", NO_NUM);
  va_end(va);
  return(m.m_type == OK ? m.RET_VALUE : m.m_type);
}
