/* getfs  -  get files from a MINIX filesystem	Authors: M. Patzel */
/* derived from A.Tanenbaum's & Paul ogilvie's mkfs */


#define _POSIX_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <ansi.h>
#include <sys/types.h>
#include <sys/dir.h>
#include <fcntl.h>
#include <unistd.h>
#include <minix/config.h>
#include <minix/const.h>
#include <minix/type.h>
#include "../fs/const.h"

#undef EXTERN
#define EXTERN			/* get rid of EXTERN by making it null */
#include "../fs/type.h"
#include "../fs/super.h"
#undef printf


#if 0
#define INODE_MAP            2
#define MAX_TOKENS          10
#define BIN                  2
#define BINGRP               2
#define BIT_MAP_SHIFT       12	/* log_2(BLOCK_SIZE) + 3 */
#define N_BLOCKS       0x10000L	/* must be multiple of 8 */
#endif

#define	READ		0
#define	BUF_SIZE	512
#define	NIL_INODE	(-1)
#define LINE_LEN           200

#define	first_inode_block (SUPER_BLOCK+sb.s_imap_blocks+sb.s_zmap_blocks+1)

typedef struct ino_file {
	d_inode	i_inode;
	off_t	i_filepos;
	struct ino_file *i_next;
} INO_FILE;
INO_FILE *if_root = NULL;

struct dsb {
  ino_t s_ninodes;		/* # inodes on the minor device */
  zone_nr s_nzones;		/* total dev size, incl. bit maps etc */
  unsigned short s_imap_blocks;	/* # of blocks used by inode bit map */
  unsigned short s_zmap_blocks;	/* # of blocks used by zone bit map */
  zone_nr s_firstdatazone;	/* number of first data zone */
  short s_log_zone_size;	/* log2 of blocks/zone */
  off_t s_maxsize;		/* maximum file size on this device */
  short s_magic;		/* magic number for super blocks */
} sb;

int fd;


_PROTOTYPE(void usage, (void)						);
_PROTOTYPE(char *unx2arm, (char *unixfn, char *armfn)			);
_PROTOTYPE(int getline, (int fd, char *unixfn, char *armfn)		);
_PROTOTYPE(int getch, (int fd)						);
_PROTOTYPE(void xfile, (char *unixfn, char *armfn)			);
_PROTOTYPE(char *ufn_rofn, (char *unixfn)				);
_PROTOTYPE(void get_super, (struct dsb *sb)				);
_PROTOTYPE(INO_FILE *open_file, (int inode_nr)				);
_PROTOTYPE(void close_file, (INO_FILE *fp)				);
_PROTOTYPE(int read_file, (INO_FILE *fp, char *buf, int n)		);
_PROTOTYPE(void get_block, (int n, char buf[BLOCK_SIZE])		);
_PROTOTYPE(void cache_init, (void)					);
_PROTOTYPE(void mx_read, (int blocknr, char buf[BLOCK_SIZE])		);
_PROTOTYPE(int inode_nr, (char *unixfn)					);
_PROTOTYPE(void get_inode, (int ino_nr, d_inode *inode)			);
_PROTOTYPE(void listdir, (int ino_nr, char *path)			);



/*================================================================
 *                    getfs  -  get files from the  filesystem
 *===============================================================*/

int
main(argc, argv)
int argc;
char *argv[];
{
  int fdw;
  char *name, unixname[LINE_LEN], armname[LINE_LEN];

  fdw = -1;
  if (argc < 2) usage();
  if (strcmp(argv[1], "-l") == 0)
  {
  	if (argc != 3) usage();
  	argv++, argc--;
  }
  if (strcmp(argv[1], "-f") == 0)
  {
	if (argc != 4) usage();
	argv += 2; argc -= 2;
	if ((fdw = open(*argv, READ)) < 0)
	{
		printf("cannot open %s\n", *argv); exit(1);
	}
  }

  if ((fd = open(*++argv, READ)) < 0)
  {
	printf("cannot open %s\n", *argv); exit(1);
  }
  argc--;

  cache_init();
  get_super(&sb);

  if (strcmp(argv[-1], "-l") == 0)
  {
	listdir(1, "");
	exit(0);
  }

  if (fdw >= 0)
  {
  	while (getline(fdw, unixname, armname))
  	{
		name = unx2arm(unixname, armname[0] ? armname : NULL);
		xfile(unixname, name);
	}
	close(fdw);
  }
  else
  {
	while (argc-- > 1)
	{
		name = unx2arm(*++argv, NULL);
		xfile(*argv, name);
	}
  }
  return(0);
}


/*================================================================
 *                 usage  -  print usage message
 *===============================================================*/

void
usage()
{
  printf("usage: getfs <drive_filename> file [files....]\n\
       getfs -f <workfile> <drive_filename>\n");
  exit(1);
}


/*================================================================
 *                 unx2arm  -  convert unixfilename to armfilename
 *===============================================================*/

char *
unx2arm(unixfn, armfn)
char *unixfn, *armfn;
{
  static char newarm[LINE_LEN];
  char line[LINE_LEN];
  char *cp, *cp1;
  FILE *fp;
  extern mkdir(char *);

  if (armfn == NULL)
  {
	cp = ufn_rofn(unixfn);
	strcpy(newarm, cp);
	free(cp);
  }
  else
  	strcpy(newarm, armfn);
  if ((fp = fopen(newarm, "r")) != NULL)
  {
	fclose(fp);
	printf("%s already exists, continue anyway? ", newarm);
	gets(line);
	if (line[0] != 'y' && line[0] != 'Y')
	{
		printf("enter new filename: ");
		gets(newarm);
	}
  }
  strcpy(line, newarm);
  cp = strrchr(line, '.');
  if (!cp)
	  return(newarm);
  *cp = '\0';
  if ((fp = fopen(line, "r")) == NULL)
  {
  	fclose(fp);
  	printf("path %s does not exist, create it? ", line);
  	gets(line);
  	if (line[0] != 'y' && line[0] != 'Y')
  		return(newarm);
  }
  for (cp = newarm, cp1 = line; *cp; )
  {
  	while (*cp && *cp != '.')
  		*cp1++ = *cp++;
	if (!*cp) break;
	*cp1 = '\0';
  	mkdir(line);
	*cp1++ = *cp++;
  }
  return(newarm);
}


/*================================================================
 *                 getline  -  read a line from workfile
 *===============================================================*/

int
getline(fd, unixfn, armfn)
int fd;
char *unixfn, *armfn;
{
  int c;
  char *cp;

  unixfn[0] = armfn[0] = '\0';
  while ((c = getch(fd)) != EOF && strchr(" \t\r\n", c))
	;
  cp = unixfn;
  while (c != EOF && !strchr(" \t\r\n", c))
	*cp++ = c, c = getch(fd);
  *cp++ = '\0';
  while (c != EOF && strchr(" \t", c))
	c = getch(fd);
  cp = armfn;
  while (c != EOF && !strchr(" \t\r\n", c))
	*cp++ = c, c = getch(fd);
  *cp++ = '\0';

  return(unixfn[0] || armfn[0] || c != EOF);
}


/*================================================================
 *                 getch  -  get one character
 *===============================================================*/

int
getch(fd)
int fd;
{
  static char buffer[BUF_SIZE], *bp;
  static char cnt = 0;

  if (cnt <= 0)
  {
	bp = buffer;
	if ((cnt = read(fd, buffer, BUF_SIZE)) <= 0)
		return (EOF);
  }
  cnt--;
  return (*bp++);
}


/*================================================================
 *                 xfile  -  extract a unix file
 *===============================================================*/

void
xfile(unixfn, armfn)
char *unixfn, *armfn;
{
  int ino_nr, n;
  d_inode inode;
  char line[20], buffer[4192];
  INO_FILE *fp;
  FILE *fpa;

  ino_nr = inode_nr(unixfn);
  if (ino_nr == NIL_INODE)
  {
	printf("%s: non existent\n", unixfn);
	return;
  }

  get_inode(ino_nr, &inode);
  if ((inode.i_mode & I_TYPE) != I_REGULAR)
  {
  	printf("%s is not a regular file, extract anyway? ");
  	gets(line);
  	if (line[0] != 'y' && line[0] != 'Y') return;
  }
  if ((fp = open_file(ino_nr)) == NULL)
  {
	printf("can't open %s for reading\n", unixfn);
	return;
  }
  if ((fpa = fopen(armfn, "wb")) == NULL)
  {
	printf("can't open %s for writing\n", armfn);
	close_file(fp);
	return;
  }
  while ((n = read_file(fp, buffer, sizeof(buffer))) > 0)
	if (fwrite(buffer, sizeof(char), n, fpa) != n)
	{
		printf("write error on %s\n", armfn);
		fclose(fpa); close_file(fp);
		return;
	}
  fclose(fpa); close_file(fp);
}


/*================================================================
 *                 listdir  -  list the file system
 *===============================================================*/

void
listdir(ino_nr, path)
int ino_nr;
char *path;
{
  INO_FILE *fp;
  d_inode inode;
  struct direct dir_ent;
  char buf[PATH_MAX];

  fp = open_file(ino_nr);
  if (fp == NULL) return;
  while (read_file(fp, (char *)&dir_ent, sizeof(dir_ent)) == sizeof(dir_ent))
  {
	if (!dir_ent.d_ino || !strcmp(dir_ent.d_name, ".")
	    || !strcmp(dir_ent.d_name, "..")) continue;
	sprintf(buf, "%s/%s", path, dir_ent.d_name);
	get_inode(dir_ent.d_ino, &inode);
	if ((inode.i_mode & I_TYPE) == I_DIRECTORY)
	{
		listdir(dir_ent.d_ino, buf);
	}
	else
		puts(buf);
  }
  close_file(fp);
}


/*===========================================================================*/
/*===========================================================================*/

/*================================================================
 *	    directory & inode management assist group
 *===============================================================*/

int
inode_nr(unix_fn)
char *unix_fn;
{
  char	name[DIRSIZ + 1], *cp, *cp1;
  int	ino_nr;
  INO_FILE *fp;
  struct direct dir_ent;

  ino_nr = 1;			/* start at root */
  cp = unix_fn;
  /* scan path and search for the appropriate inode number */
  while (*cp)
  {
	/* extract next part of the path */
	if (*cp == '/') { cp++; continue; }
	cp1 = name;
	while (*cp !=  '/' && *cp && cp1 - name < DIRSIZ)
		*cp1++ = *cp++;
	*cp1 = '\0';

	/* search a part of the path in the current (ino_nr) directory */
	fp = open_file(ino_nr);
	if (fp == NULL)
	{
		printf("can't open inode %d in path %s (part=%s)\n",
			ino_nr, unix_fn, name);
		return(NIL_INODE);
	}
	while (TRUE)
	{
		/* read one directory entry */
		if (   read_file(fp, (char *)&dir_ent, sizeof(dir_ent))
		    != sizeof(dir_ent))
		{
			close_file(fp);
			printf("can't find part %s in path %s\n",
				name, unix_fn);
			return(NIL_INODE);
		}
		if (strcmp(name, dir_ent.d_name) == 0)
		{
			ino_nr = dir_ent.d_ino;
			close_file(fp);
			break;
		}
	}			
  }
  return(ino_nr);
}

void
get_inode(ino_nr, inode)
int ino_nr;
d_inode *inode;
{
  d_inode ino[INODES_PER_BLOCK];

  ino_nr--;
  get_block(first_inode_block + ino_nr / INODES_PER_BLOCK, (char *)ino);
  memcpy(inode, &ino[ino_nr % INODES_PER_BLOCK], sizeof(d_inode));
}

void
get_super(sb)
struct dsb *sb;
{
  char buf[BLOCK_SIZE];

  get_block(SUPER_BLOCK, buf);
  memcpy(sb, buf, sizeof(struct dsb));
}


/*===========================================================================*/
/*===========================================================================*/

/*================================================================
 *	    file management group
 *===============================================================*/

INO_FILE *
open_file(inode_nr)
int inode_nr;
{
  INO_FILE *ifp;

  if ((ifp = malloc(sizeof(INO_FILE))) == NULL)
	return(NULL);
  get_inode(inode_nr, &ifp->i_inode);
  ifp->i_next = if_root; if_root = ifp;
  ifp->i_filepos = 0;
  return(ifp);
}

void
close_file(fp)
INO_FILE *fp;
{
  INO_FILE *ip;

  if (fp == if_root)
	if_root = fp->i_next;
  else
  {
	ip = if_root;
	while (ip && ip->i_next != fp) ip = ip->i_next;
	if (ip)
		ip->i_next = fp->i_next;
  }
  free(fp);
}

int
read_file(fp, buf, n)
INO_FILE *fp;
char *buf;
int n;
{
  static char data[BLOCK_SIZE];
  static zone_nr blk[NR_INDIRECTS], dblk[NR_INDIRECTS];
  static int blknum = -1, dblknum = -1;
  int cnt, sum = 0, b, b1;

  if (fp->i_filepos + n > fp->i_inode.i_size)
	n = (int)(fp->i_inode.i_size - fp->i_filepos);
  while (n > 0)
  {
	b = (int)(fp->i_filepos / BLOCK_SIZE);
	if (b < NR_DZONE_NUM)
		get_block(fp->i_inode.i_zone[b], data);
	else if (b < NR_DZONE_NUM + NR_INDIRECTS)
	{
		if (blknum != fp->i_inode.i_zone[NR_DZONE_NUM])
			get_block(blknum = fp->i_inode.i_zone[NR_DZONE_NUM],
				  (char *)blk);
		get_block(blk[b - NR_DZONE_NUM], data);
	}
	else
	{
		b1 = (b - NR_DZONE_NUM - NR_INDIRECTS);
		if (dblknum != fp->i_inode.i_zone[NR_DZONE_NUM+1])
			get_block(dblknum = fp->i_inode.i_zone[NR_DZONE_NUM+1],
				  (char *)dblk);
		if (blknum != dblk[b1 / NR_INDIRECTS])
			get_block(blknum = dblk[b1 / NR_INDIRECTS], (char *)blk);
		get_block(blk[b1 % NR_INDIRECTS], data);
	}
	if (fp->i_filepos % BLOCK_SIZE + n <= BLOCK_SIZE)
		cnt = n;
	else
		cnt = (int)(BLOCK_SIZE - fp->i_filepos % BLOCK_SIZE);
	memcpy(buf, data + fp->i_filepos % BLOCK_SIZE, cnt);
	fp->i_filepos += cnt; sum += cnt; buf += cnt;
	n -= cnt;
  }
  return(sum);
}

/*===========================================================================*/
/*===========================================================================*/

/*================================================================
 *	    block management group
 *===============================================================*/

char *derrtab[14] = {
	     "no error",
	     "disk is read-only",
	     "unknown unit",
	     "device not ready",
	     "bad command",
	     "data error",
	     "internal error: bad request structure length",
	     "seek error",
	     "unknown media type",
	     "sector not found",
	     "printer out of paper",
	     "write fault",
	     "read error",
	     "general error"
};

#define	CACHE_SIZE	100	/* 100 block-buffers */
struct cache {
  char *blockbuf;
  int blocknum;
  int dirty;
  int usecnt;
} cache[CACHE_SIZE];


void
get_block(n, buf)
int n;
char buf[BLOCK_SIZE];
{
  /* Get a block to the user */
  struct cache *bp, *fp;

  /* Look for block in cache */
  fp = 0;
  for (bp = cache; bp < &cache[CACHE_SIZE]; bp++) {
	if (bp->blocknum == n) {
		memcpy(buf, bp->blockbuf, BLOCK_SIZE);
		bp->usecnt++;
		return;
	}

	/* Remember clean block */
	if (bp->dirty == 0)
	{
		if (fp) {
			if (fp->usecnt > bp->usecnt) fp = bp;
		} else
			fp = bp;
	}
  }

  mx_read(n, fp->blockbuf);
  fp->dirty = 0;
  fp->usecnt = 0;
  fp->blocknum = n;
  memcpy(buf, fp->blockbuf, BLOCK_SIZE);
}

void
cache_init()
{
  struct cache *bp;
  for (bp = cache; bp < &cache[CACHE_SIZE]; bp++)
  {
	bp->blocknum = -1;
	if ((bp->blockbuf = malloc(BLOCK_SIZE)) == NULL)
		perror("blockbuf");
  }
}

void
mx_read(blocknr, buf)
int blocknr;
char buf[BLOCK_SIZE];
{
  lseek(fd, (long) blocknr * BLOCK_SIZE, 0);
  if (read(fd, buf, BLOCK_SIZE) != BLOCK_SIZE)
  {
	printf("mx_read: error reading file");
	exit(1);
  }
}


/*===========================================================================*/
/*===========================================================================*/

/*----------------------------------------------------*
 * Module : c.ufn_rofn
 * Purpose: Convert UNIX filenames to RiscOS filename
 * Author : Steven van Aardt
 *----------------------------------------------------*/

#define MAX_FILENAME 10

/*---------------------------- Reverse ------------------------*
 * Purpose:   Reverse the character sequence within a string
 * Arguments: Pointer to string
 * Returns:   Pointer to the same string given as an argument.
 *-------------------------------------------------------------*/
static
char *reverse(char *string)
{
  int hd=0,tl=strlen(string);
  char tmp_c;
  /* reverse the name */
  while (--tl>hd) {                   /* work way back from end of string */
    tmp_c=string[hd];
    string[hd++]=string[tl];
    string[tl]=tmp_c;
  }
  return string;
}

/*---------------------------- isvowel --------------------------------*
 * Purpose:   Determines whether a character is an alphabetical vowel
 * Arguments: character (integer form)
 * Returns:   TRUE if vowel, otherwise FALSE.
 *---------------------------------------------------------------------*/
static
int isvowel(int c)
{
  int vowel=FALSE;
  switch (c) {
    case 'a': case 'A':             /* if c matches with */
    case 'e': case 'E':             /* any of these function */
    case 'i': case 'I':             /* will return TRUE */
    case 'o': case 'O':
    case 'u': case 'U': vowel=TRUE;
  }
  return vowel;
}

/*--------------------  squeeze  ------------------------*
 * Purpose:   remove any space characters from a string
 * Arguments: pointer to string
 * Returns:   pointer to a new string.
 *-------------------------------------------------------*/
static
char *squeeze(char *ptr)
{
   int  i=0,j=0;
   char *nn=NULL;
   /* squeeze in-place */
   for (i=j=0; ptr[i]!='\0';i++) if (ptr[i]!=' ') ptr[j++]=ptr[i];
   /* and copy into a new string */
   nn=calloc(j+2,sizeof(char)); /* allocate enough memory for new string */
   strncpy(nn,ptr,j);
   return nn;
}

/*---------------------------- rm_vowels -----------------------------*
 * Purpose:   removes some vowels from a string thus shortening it
 * Arguments: pointer to string
 * Returns:   pointer to a new string.
 *--------------------------------------------------------------------*/
static
char *rm_vowels(char *word,int maxlen)
{
  char *new_word=NULL;
  if (word) {
    int len=strlen(word);
    char *cpy=calloc(len+1,sizeof(char));
    strcpy(cpy,word);
    if (len>maxlen) { 
      /* remove every 2nd vowel */
      int i;
      len--;  /* leave last letter */
      for (i=1; i<len; i+=2) {
        while ((i<len) && (!isvowel(cpy[i]))) i++; /* skip non-vowels */ 
        if ((i<len) && (isvowel(cpy[i]))) cpy[i]=' '; /* overwrite vowels */
      }
    }
    new_word=squeeze(cpy); 
    free(cpy);
  }
  return new_word;
}

/*-------------------  shorten  ----------------------*
 * Purpose:   Shortens a string to a specified length
 * Arguments: pointer to string
 *            max length of string to be.
 * Returns: Pointer to new string
 *----------------------------------------------------*/
static
char *shorten(char *ptr,int maxlen)
{
  char *new_ptr;
  int l=strlen(ptr);
  if (l>maxlen) {
    /* quick shortening by removing a chunk from the middle */
    int num=l-maxlen; /* number to remove to get length of 10 */
    int i,j=0;
    for (i=(l-num)/2; j<num; j++) ptr[i+j]=' ';
  }
  new_ptr=squeeze(ptr);
  return new_ptr;
}



/*--------------------- ufn_rofn ------------------------*
 * Purpose:   Convert UNIX filename to RISC OS filename 
 * Arguments: pointer to UNIX filename string
 * Returns:   pointer to RISC OS filename
 *-------------------------------------------------------*/
char *ufn_rofn(char *unix_name)
{
  typedef struct l List;         /* structure for constituent parts */
  struct l { char *ptr; struct l *last,*next; };
  int  len;
  char *uname=calloc(strlen(unix_name)+1,sizeof(char)),*rname=NULL,*ptr=NULL;
  List *lptr=NULL,*hdptr=NULL,*tlptr=NULL;
  while ((unix_name) && (*unix_name==' ')) unix_name++;
  if ((unix_name) && (*unix_name)) {
    strcpy(uname,unix_name);
    /* look for root directory '/' at beginning of unix filename */
    if (uname[0]=='/') {
      char *tmp=calloc(2,sizeof(char));
      strcpy(tmp,"$");
      lptr=(List *) malloc(sizeof(List));
      tlptr=lptr;         /* pointer to beginning of list */
      lptr->ptr=tmp;      
      lptr->next=NULL;
      lptr->last=hdptr;
      hdptr=lptr;         /* pointer to end of list */
    }
   /*
    *   Break down the name into its directory components - destroys uname
    */
    ptr=strtok(uname,"/");          /* ignores / if at beginning of list */
    while (ptr) {
      char *tmp=calloc(strlen(ptr)+1,sizeof(char));
      strcpy(tmp,ptr);
      lptr=(List *) malloc(sizeof(List));
      if (!tlptr) tlptr=lptr;      /* make pointer to beginning of list */
      if (hdptr) hdptr->next=lptr; /* link previous element onto this */
      lptr->ptr=tmp;
      lptr->next=NULL;   
      lptr->last=hdptr;  /* reverse chained list */
      hdptr=lptr;
      ptr=strtok(NULL,"/");        /* find next '/' */
    }
    /*
     *  Now go through each component looking for a '.' within each.
     */
    lptr=tlptr;                     /* beginning of list */
    while (lptr) {
      char *dp=NULL;                /* dot position */
      dp=strchr(lptr->ptr,'.');
      if (dp) {
        int len=strlen(lptr->ptr);
        /* 
         * special cases
         */
        /* if only 1 dot then refers to the UNIX csd */
        if (len==1) { 
          char *tmp=calloc(2,sizeof(char));
          strcpy(tmp,"@");
          lptr->ptr=tmp;
        } else {
          if (len==2) { 
            char *tmp=calloc(2,sizeof(char)); 
            strcpy(tmp,"^");
            lptr->ptr=tmp;
          } else {
            /* 
             * General case requires reversing of parts around '.'
             */
            List *dlptr=NULL, *dhdptr=NULL, *dtlptr=NULL,*tmp_next=lptr->next;
            char *dptr=NULL,*cpy=calloc(len+1,sizeof(char));

            strcpy(cpy,lptr->ptr);
            reverse(cpy);           /* reverse complete text */
            dptr=strtok(cpy,".");
            while (dptr) {
              char *tmp=calloc(strlen(dptr)+1,sizeof(char));  
              strcpy(tmp,dptr); /* make a copy of part */ 
              dlptr=(List *) malloc(sizeof(List));
              if (!dtlptr) dtlptr=dlptr;  /* remember beginning of this list */
              if (dhdptr) dhdptr->next=dlptr; /* link previous one onto this */
              dlptr->ptr=reverse(tmp); /* and reverse it's text back again */
              dlptr->next=NULL;
              dlptr->last=dhdptr;  /* reverse chaining  */
              dhdptr=dlptr;
              dptr=strtok(NULL,".");
            } /* endwhile */
            if (lptr->last) {
              free(lptr->ptr);       /* free text worked on by above */
              lptr=lptr->last;       /* move back */
              free(lptr->next);      /* free text's list element */
              lptr->next=dtlptr;     /* link */
              dtlptr->last=lptr;
              dlptr->next=tmp_next;
              if (tmp_next) { 
                tmp_next->last=dlptr;
                lptr=tmp_next->last;
              } else lptr=dlptr;
            } else {        /* case when this is the head of the list */
              free(lptr->ptr);
              free(lptr);
              lptr=dtlptr;
              tlptr=lptr;
              lptr->last=NULL;
              dlptr->next=tmp_next;
              if (tmp_next) {
                tmp_next->last=dlptr;
                lptr=tmp_next->last;
              } else lptr=dlptr;
            } /* endif */
          } /* endif */
        } /* endif */
      } /* endif */ 
      lptr=lptr->next; /* move onto next directory component in list */
    } /* endwhile */
    /* 
     *  Check the lengths of each component and adjust to that for RISC OS
     */
    lptr=tlptr;
    len=0;
    while (lptr) {
      char *new_ptr=NULL;
      if (lptr->next) len+=1;  /* for the RISC OS directory separator */
      new_ptr=rm_vowels(lptr->ptr,MAX_FILENAME);
      free(lptr->ptr);
      lptr->ptr=new_ptr;
      new_ptr=shorten(lptr->ptr,MAX_FILENAME);
      free(lptr->ptr);
      lptr->ptr=new_ptr;
      len+=strlen(lptr->ptr);
      lptr=lptr->next;
    }
    /*
     *  Copy linked list into result string
     */
    rname=calloc(len+1,sizeof(char)); /* make suitable string space */
    lptr=tlptr;
    len=0;
    while (lptr) {
      strcat((rname+len),lptr->ptr);
      free(lptr->ptr);     
      len=strlen(rname);
      lptr=lptr->next;
      if (lptr) strcat(rname,"."); /* append RISC OS directory separator */
    }
    /*
     *  Tidy up, by freeing memory used by list
     */
    lptr=tlptr;
    while (lptr) {
      List *tmp;
      tmp=lptr;
      lptr=lptr->next;
      free(tmp);
    }
  } /* endif */
  free(uname);
  return rname;
}
